@@ -20,6 +20,7 @@ from openpiv.pyprocess import extended_search_area_piv, get_coordinates, \
     get_field_shape
 from openpiv import smoothn
 from skimage.measure import points_in_poly
+from skimage.util import invert
 
 
 def piv(settings):
@@ -61,14 +62,19 @@ def piv(settings):
                 settings.ROI[0]:settings.ROI[1],
                 settings.ROI[2]:settings.ROI[3]
             ]
-        if settings.dynamic_masking_method in ("edge","intensity"):
+
+        if hasattr(settings, 'invert') and settings.invert is True:
+            frame_a = invert(frame_a)
+            frame_b = invert(frame_b)
+
+        if settings.dynamic_masking_method in ("edge", "intensity"):
             frame_a = preprocess.dynamic_masking(
                 frame_a,
                 method=settings.dynamic_masking_method,
                 filter_size=settings.dynamic_masking_filter_size,
                 threshold=settings.dynamic_masking_threshold,
             )
-            frame_b = preprocess.dynamic_masking(
+            frame_b, mask_b = preprocess.dynamic_masking(
                 frame_b,
                 method=settings.dynamic_masking_method,
                 filter_size=settings.dynamic_masking_filter_size,
@@ -76,12 +82,12 @@ def piv(settings):
             )
 
         # "first pass"
-        x, y, u, v, sig2noise_ratio = first_pass(
+        x, y, u, v, s2n = first_pass(
             frame_a,
             frame_b,
             settings.windowsizes[0],
             settings.overlap[0],
-            settings.iterations,
+            settings.num_iterations,
             correlation_method=settings.correlation_method,
             subpixel_method=settings.subpixel_method,
             do_sig2noise=settings.extract_sig2noise,
@@ -90,8 +96,21 @@ def piv(settings):
             normalized_correlation=settings.normalized_correlation
         )
 
-        "validation using gloabl limits and std and local median"
-        """MinMaxU : two elements tuple
+        # " Image masking "
+        if hasattr(settings, 'image_mask') and settings.image_mask:
+            image_mask = np.logical_and(mask_a, mask_b)
+            mask_coords = preprocess.mask_coordinates(image_mask)
+            # mark those points on the grid of PIV inside the mask
+            xymask = points_in_poly(np.c_[y.flatten(), x.flatten()], mask_coords)
+            tmp = np.zeros_like(x,dtype=bool)
+            tmp.flat[xymask] = True
+            u = np.ma.masked_array(u, mask=tmp)
+            v = np.ma.masked_array(v, mask=tmp)
+
+        """
+        validation using gloabl limits and std and local median
+
+        MinMaxU : two elements tuple
             sets the limits of the u displacment component
             Used for validation.
 
@@ -116,7 +135,8 @@ def piv(settings):
             maximum of filter iterations to replace nans
 
         filter_kernel_size : int
-            size of the kernel used for the filtering"""
+            size of the kernel used for the filtering
+        """
 
         mask = np.full_like(x, False)
         if settings.validation_first_pass is True:
@@ -133,20 +153,24 @@ def piv(settings):
                 v_threshold=settings.median_threshold,
                 size=settings.median_size,
             )
+            mask = mask + mask_g + mask_m + mask_s
+
             if (
                 settings.extract_sig2noise is True
-                and settings.iterations == 1
+                and settings.num_iterations == 1
                 and settings.do_sig2noise_validation is True
             ):
                 u, v, mask_s2n = validation.sig2noise_val(
-                    u, v, sig2noise_ratio,
+                    u, v, s2n,
                     threshold=settings.sig2noise_threshold
                 )
-                mask = mask + mask_g + mask_m + mask_s + mask_s2n
-            else:
-                mask = mask + mask_g + mask_m + mask_s
-        "filter to replace the values that where marked by the validation"
-        if settings.iterations > 1:
+                mask += mask_s2n
+                
+        
+        # "filter to replace the values that where marked by the validation"
+        if settings.num_iterations == 1 and settings.replace_vectors is True:
+            # for multi-pass we cannot have holes in the data
+            # after the first pass
             u, v = filters.replace_outliers(
                 u,
                 v,
@@ -154,15 +178,7 @@ def piv(settings):
                 max_iter=settings.max_filter_iteration,
                 kernel_size=settings.filter_kernel_size,
             )
-            "adding masks to add the effect of all the validations"
-            if settings.smoothn is True:
-                u, dummy_u1, dummy_u2, dummy_u3 = smoothn.smoothn(
-                    u, s=settings.smoothn_p
-                )
-                v, dummy_v1, dummy_v2, dummy_v3 = smoothn.smoothn(
-                    v, s=settings.smoothn_p
-                )
-        elif settings.iterations == 1 and settings.replace_vectors is True:
+        elif settings.num_iterations > 1: # don't even check if it's true or false
             u, v = filters.replace_outliers(
                 u,
                 v,
@@ -170,30 +186,31 @@ def piv(settings):
                 max_iter=settings.max_filter_iteration,
                 kernel_size=settings.filter_kernel_size,
             )
-            "adding masks to add the effect of all the validations"
-            if settings.smoothn is True:
-                u, v = filters.replace_outliers(
-                    u,
-                    v,
-                    method=settings.filter_method,
-                    max_iter=settings.max_filter_iteration,
-                    kernel_size=settings.filter_kernel_size,
-                )
-                u, dummy_u1, dummy_u2, dummy_u3 = smoothn.smoothn(
-                    u, s=settings.smoothn_p
-                )
-                v, dummy_v1, dummy_v2, dummy_v3 = smoothn.smoothn(
-                    v, s=settings.smoothn_p
-                )
 
-        # "all the following passes"
-        for i in range(1, settings.iterations):
-            x, y, u, v, sig2noise_ratio = multipass_img_deform(
+            # "adding masks to add the effect of all the validations"
+        if settings.smoothn is True:
+            u, dummy_u1, dummy_u2, dummy_u3 = smoothn.smoothn(
+                u, s=settings.smoothn_p
+            )
+            v, dummy_v1, dummy_v2, dummy_v3 = smoothn.smoothn(
+                v, s=settings.smoothn_p
+            )
+
+        # plt.figure()
+        # plt.quiver(x,y,u,v)
+        # plt.gca().invert_yaxis()
+        # plt.title('after first pass')
+        # plt.show()
+    
+        """ Multi pass """
+        
+        for i in range(1, settings.num_iterations):
+            x, y, u, v, s2n = multipass_img_deform(
                 frame_a,
                 frame_b,
                 settings.windowsizes[i],
                 settings.overlap[i],
-                settings.iterations,
+                settings.num_iterations,
                 i,
                 x,
                 y,
@@ -208,7 +225,8 @@ def piv(settings):
                 interpolation_order=settings.interpolation_order,
                 normalized_correlation=settings.normalized_correlation
             )
-            mask = np.full_like(x,False,dtype=bool)
+
+            mask = np.full_like(x, False, dtype=bool)
             # validation every step
             u, v, mask_g = validation.global_val(
                 u, v, settings.MinMax_U_disp, settings.MinMax_V_disp
@@ -224,14 +242,33 @@ def piv(settings):
                 size=settings.median_size,
             )
             mask = mask + mask_s + mask_m + mask_g
+
+            if settings.do_sig2noise_validation is True:
+
+                u, v, mask_s2n = validation.sig2noise_val(
+                    u, v, s2n, threshold=settings.sig2noise_threshold
+                )
+                mask = mask + mask_s2n
+            
+            # we have to replace outliers
+            u, v = filters.replace_outliers(
+                u,
+                v,
+                method=settings.filter_method,
+                max_iter=settings.max_filter_iteration,
+                kernel_size=settings.filter_kernel_size,
+            )
+
             # If the smoothing is active, we do it at each pass
-            if settings.smoothn is True:
+            # but not the last one
+            if settings.smoothn is True and i < settings.num_iterations-1: 
                 u, dummy_u1, dummy_u2, dummy_u3 = smoothn.smoothn(
                     u, s=settings.smoothn_p
                 )
                 v, dummy_v1, dummy_v2, dummy_v3 = smoothn.smoothn(
                     v, s=settings.smoothn_p
                 )
+
             if (
                 settings.extract_sig2noise is True
                 and i == settings.iterations
@@ -239,61 +276,61 @@ def piv(settings):
                 and settings.do_sig2noise_validation is True
             ):
                 u, v, mask_s2n = validation.sig2noise_val(
-                    u, v, sig2noise_ratio, threshold=settings.sig2noise_threshold
+                    u, v, sig2noise_ratio,
+                    threshold=settings.sig2noise_threshold
                 )
                 mask = mask + mask_s2n
 
-            
             # replace also every loop
-            if settings.replace_vectors is True:
-                u, v = filters.replace_outliers(
-                    u,
-                    v,
-                    method=settings.filter_method,
-                    max_iter=settings.max_filter_iteration,
-                    kernel_size=settings.filter_kernel_size,
-                )
+            u, v = filters.replace_outliers(
+                u,
+                v,
+                method=settings.filter_method,
+                max_iter=settings.max_filter_iteration,
+                kernel_size=settings.filter_kernel_size,
+            )
 
-        "pixel/frame->pixel/sec"
+        # "pixel/frame->pixel/sec"
         u = u / settings.dt
         v = v / settings.dt
-        "scales the results pixel-> meter"
+        
+        # "scales the results pixel-> meter"
         x, y, u, v = scaling.uniform(x, y, u, v,
                                      scaling_factor=settings.scaling_factor)
-        "save to a file"
+        
+        # "save to a file"
         save(
             x,
             y,
             u,
             v,
-            sig2noise_ratio,
+            s2n,
             mask,
             os.path.join(save_path, "field_A%03d.txt" % counter),
             delimiter="\t",
         )
-        "some messages to check if it is still alive"
-
-        "some other stuff that one might want to use"
+        
+        # "some other stuff that one might want to use"
         if settings.show_plot is True or settings.save_plot is True:
             plt.close("all")
             plt.ioff()
             Name = os.path.join(save_path, "Image_A%03d.png" % counter)
-            display_vector_field(
+            fig, _ = display_vector_field(
                 os.path.join(save_path, "field_A%03d.txt" % counter),
                 scale=settings.scale_plot,
             )
             if settings.save_plot is True:
-                plt.savefig(Name)
+                fig.savefig(Name)
             if settings.show_plot is True:
                 plt.show()
 
         print("Image Pair " + str(counter + 1))
 
-    "Below is code to read files and create a folder to store the results"
+    # "Below is code to read files and create a folder to store the results"
     save_path = os.path.join(
         settings.save_path,
         "Open_PIV_results_"
-        + str(settings.windowsizes[settings.iterations-1])
+        + str(settings.windowsizes[settings.num_iterations-1])
         + "_"
         + settings.save_folder_suffix,
     )
@@ -510,6 +547,7 @@ def first_pass(
 
     return x, y, u, v, s2n
 
+
 def multipass_img_deform(
     frame_a,
     frame_b,
@@ -577,7 +615,7 @@ def multipass_img_deform(
     interpolation_order : int
         the order of the spline interpolation used for the image deformation
 
-    mask_coords : list of x,y coordinates (pixels) of the image mask, 
+    mask_coords : list of x,y coordinates (pixels) of the image mask,
         default is an empty list
 
     Returns
@@ -589,43 +627,40 @@ def multipass_img_deform(
         array containg the y coordinates of the interrogation window centres
 
     u : 2d np.array
-        array containing the horizontal displacement for every interrogation window
-        [pixels]
+        array containing the horizontal displacement for every interrogation
+        window [pixels]
 
     u : 2d np.array
-        array containing the vertical displacement for every interrogation window
-        it returns values in [pixels]
-        
+        array containing the vertical displacement for every interrogation
+        window it returns values in [pixels]
+
     s2n : 2D np.array of signal to noise ratio values
 
     """
 
     y_old = y_old[:, 0]
     x_old = x_old[0, :]
-    
-    
-    # calculate the y and y coordinates of the interrogation window centres. 
+
+    # calculate the y and y coordinates of the interrogation window centres.
     # The interpolation function dont like meshgrids as input. Hence, the
     # edges must be extracted to provide the sufficient input. x_old and y_old
     # are the coordinates of the old grid. x_int and y_int are the coordinates
     # of the new grid
-    
-    x, y = get_coordinates(np.shape(frame_a), 
-                           window_size, 
+
+    x, y = get_coordinates(np.shape(frame_a),
+                           window_size,
                            overlap)
 
-
-
     # reapply the image mask to the new coordinates
     if len(mask_coords) > 1:  # not an empty list means there is a mask
         xymask = points_in_poly(np.c_[y.flatten(), x.flatten()], mask_coords)
-        mask = np.zeros_like(x,dtype=bool)
+        mask = np.zeros_like(x, dtype=bool)
         mask.flat[xymask] = 1
 
-
     y_int = y[:, 0]
     x_int = x[0, :]
 
+
     # interpolating the displacements from the old grid onto the new grid
     # y befor x because of numpy works row major
     ip = RectBivariateSpline(y_old, x_old, u_old, kx=2, ky=2)
@@ -641,12 +676,13 @@ def multipass_img_deform(
     # if settings.show_plot:
     #     plt.figure()
     #     plt.quiver(x_old, y_old, u_old, v_old,color='b')
-    #     plt.quiver(x_int, y_int, u_pre, v_pre,color='r')
+    #     plt.quiver(x_int, y_int, u_pre, v_pre,color='r',lw=2)
     #     plt.gca().invert_yaxis()
+    #     plt.title('inside deform')
     #     plt.gca().set_aspect(1.)
 
-    # @TKauefer added another method to the windowdeformation, 'symmetric' 
-    # splits the onto both frames, takes more effort due to additional 
+    # @TKauefer added another method to the windowdeformation, 'symmetric'
+    # splits the onto both frames, takes more effort due to additional
     # interpolation however should deliver better results
 
     # old_frame_a = frame_a.copy()
@@ -676,7 +712,7 @@ def multipass_img_deform(
     #     plt.imshow(frame_b-old_frame_b)
 
     if (
-        do_sig2noise is True and  # and also the last one is optional? 
+        do_sig2noise is True and  # and also the last one is optional?
         current_iteration == iterations and  # so only the last one is applied?
         iterations != 1  # we should not call multipass at all with 1 iteration
     ):
@@ -697,8 +733,8 @@ def multipass_img_deform(
         normalized_correlation=normalized_correlation,
     )
 
-    shapes = np.array(get_field_shape(frame_a.shape, 
-                                      window_size, 
+    shapes = np.array(get_field_shape(frame_a.shape,
+                                      window_size,
                                       overlap))
     u = u.reshape(shapes)
     v = v.reshape(shapes)
@@ -724,12 +760,131 @@ def multipass_img_deform(
     #     plt.quiver(x_int, y_int, u_pre, v_pre,color='b')
     #     plt.gca().invert_yaxis()
     #     plt.gca().set_aspect(1)
-    
+
     return x, y, u, v, s2n
 
 
 class Settings(object):
-    pass
+    def __init__(self):
+        # "Data related settings"
+        # Folder with the images to process
+        self.filepath_images = "."
+        # Folder for the outputs
+        self.save_path = "./res"
+        # Root name of the output Folder for Result Files
+        self.save_folder_suffix = "Test_4"
+        # Format and Image Sequence
+        self.frame_pattern_a = 'exp1_001_a.bmp'
+        self.frame_pattern_b = 'exp1_001_b.bmp'
+
+        # "Region of interest"
+        # (50,300,50,300) #Region of interest: (xmin,xmax,ymin,ymax) or 'full'
+        # for full image
+        self.ROI = "full"
+
+        # "Image preprocessing"
+        # 'None' for no masking, 'edges' for edges masking, 'intensity' for
+        # intensity masking
+        # WARNING: This part is under development so better not to use MASKS
+        self.dynamic_masking_method = "None"
+        self.dynamic_masking_threshold = 0.005
+        self.dynamic_masking_filter_size = 7
+
+        # "Processing Parameters"
+        self.correlation_method = "circular"  # 'circular' or 'linear'
+        self.normalized_correlation = False
+
+        # add the interroagtion window size for each pass.
+        # For the moment, it should be a power of 2
+        self.windowsizes = (
+            64,
+            32,
+            16,
+        )  # if longer than n iteration the rest is ignored
+        # The overlap of the interroagtion window for each pass.
+        self.overlap = (32, 16, 8)  # This is 50% overlap
+        # Has to be a value with base two. In general window size/2 is a good
+        # choice.
+
+        self.iterations = len(self.windowsizes)  # select the number of PIV
+        # passes
+
+        # methode used for subpixel interpolation:
+        # 'gaussian','centroid','parabolic'
+        self.subpixel_method = "gaussian"
+        # 'symmetric' or 'second image', 'symmetric' splits the deformation
+        # both images, while 'second image' does only deform the second image.
+        self.deformation_method = 'symmetric'  # 'symmetric' or 'second image'
+        # order of the image interpolation for the window deformation
+        self.interpolation_order = 3
+        self.scaling_factor = 1  # scaling factor pixel/meter
+        self.dt = 1  # time between to frames (in seconds)
+
+        # "Signal to noise ratio options (only for the last pass)"
+        # It is possible to decide if the S/N should be computed (for the last
+        # pass) or not
+        # 'True' or 'False' (only for the last pass)
+        self.extract_sig2noise = False
+        # method used to calculate the signal to noise ratio 'peak2peak' or
+        # 'peak2mean'
+        self.sig2noise_method = "peak2peak"
+        # select the width of the masked to masked out pixels next to the main
+        # peak
+        self.sig2noise_mask = 2
+        # If extract_sig2noise==False the values in the signal to noise ratio
+        # output column are set to NaN
+
+        # "vector validation options"
+        # choose if you want to do validation of the first pass: True or False
+        self.validation_first_pass = True
+        # only effecting the first pass of the interrogation the following
+        # passes
+        # in the multipass will be validated
+
+        # "Validation Parameters"
+        # The validation is done at each iteration based on three filters.
+        # The first filter is based on the min/max ranges. Observe that these
+        # values are defined in
+        # terms of minimum and maximum displacement in pixel/frames.
+        self.MinMax_U_disp = (-30, 30)
+        self.MinMax_V_disp = (-30, 30)
+        # The second filter is based on the global STD threshold
+        self.std_threshold = 10  # threshold of the std validation
+        # The third filter is the median test (not normalized at the moment)
+        self.median_threshold = 3  # threshold of the median validation
+        # On the last iteration, an additional validation can be done based on
+        # the S/N.
+        self.median_size = 1  # defines the size of the local median
+
+        # "Validation based on the signal to noise ratio"
+        # Note: only available when extract_sig2noise==True and only for the
+        # last pass of the interrogation
+        # Enable the signal to noise ratio validation. Options: True or False
+        self.do_sig2noise_validation = False  # This is time consuming
+        # minmum signal to noise ratio that is need for a valid vector
+        self.sig2noise_threshold = 1.05
+
+        # "Outlier replacement or Smoothing options"
+        # Replacment options for vectors which are masked as invalid by the
+        # validation
+        # Choose: True or False
+        self.replace_vectors = True  # Enable the replacement.
+        self.smoothn = True  # Enables smoothing of the displacement field
+        self.smoothn_p = 0.5  # This is a smoothing parameter
+        # select a method to replace the outliers:
+        # 'localmean', 'disk', 'distance'
+        self.filter_method = "localmean"
+        # maximum iterations performed to replace the outliers
+        self.max_filter_iteration = 4
+        self.filter_kernel_size = 2  # kernel size for the localmean method
+
+        # "Output options"
+        # Select if you want to save the plotted vectorfield: True or False
+        self.save_plot = True
+        # Choose wether you want to see the vectorfield or not :True or False
+        self.show_plot = True
+        self.scale_plot = 100  # select a value to scale the quiver plot of
+        # the vectorfield run the script with the given settings
 
 
 if __name__ == "__main__":
@@ -768,7 +923,7 @@ if __name__ == "__main__":
     "Processing Parameters"
     settings.correlation_method = "circular"  # 'circular' or 'linear'
     settings.normalized_correlation = False
-    settings.iterations = 3  # select the number of PIV passes
+    settings.num_iterations = 3  # select the number of PIV passes
     # add the interroagtion window size for each pass.
     # For the moment, it should be a power of 2
     settings.windowsizes = (
